import sys

if sys.platform == "win32":
    from rerain.functions.sys.audio import mute, unmute
    from rerain.functions.sys.mic import micmute, micunmute, record
else:
    mute = None
    unmute = None

from rerain.functions.math.math import add
from rerain.functions.math.math import sub
from rerain.functions.math.math import mult
from rerain.functions.math.math import div
from rerain.functions.math.math import power
from rerain.functions.math.math import sqrt
from rerain.functions.math.math import abs_value
from rerain.functions.math.math import log
from rerain.functions.math.math import log10
from rerain.functions.math.math import sin
from rerain.functions.math.math import cos
from rerain.functions.math.math import tan
from rerain.functions.math.math import atan
from rerain.functions.math.math import nth_root
from rerain.functions.math.math import factorial
from rerain.functions.math.math import exp
from rerain.functions.math.math import pi
from rerain.functions.math.math import e
from rerain.functions.math.math import sin_deg
from rerain.functions.math.math import cos_deg
from rerain.functions.math.math import tan_deg
from rerain.functions.math.math import derivative
from rerain.functions.math.math import integral
from rerain.functions.math.math import gcd
from rerain.functions.math.math import lcm
from rerain.functions.math.math import combination
from rerain.functions.math.math import permutation
from rerain.functions.qr.qr import newqr
from rerain.functions.sys.sys import find
from rerain.functions.sys.sys import findf
from rerain.functions.sys.sys import findall
from rerain.functions.sys.sys import create
from rerain.functions.sys.sys import rm
from rerain.functions.sys.time import wait
from rerain.functions.sys.time import wait_seconds
from rerain.functions.sys.time import wait_minutes
from rerain.functions.sys.time import wait_hours
from rerain.functions.sys.time import ct

__all__ = [
    "add", "sub", "mult", "div", "power", "sqrt", "abs_value", "log", "log10",
    "sin", "cos", "tan", "atan", "nth_root", "factorial", "exp", "pi", "e",
    "sin_deg", "cos_deg", "tan_deg", "derivative", "integral", "gcd", "lcm",
    "combination", "permutation", "newqr", "find", "findf", "findall", "create",
    "rm", "wait", "wait_seconds", "wait_minutes", "wait_hours", "ct", "micmute", "micunmute", "record"
]

if sys.platform == "win32":
    __all__.append("mute")
    __all__.append("unmute")
    __all__.append("micmute")
    __all__.append("micunmute")
    __all__.append("record")


import time

def format_time(seconds):
    minutes = seconds // 60
    hours = minutes // 60
    days = hours // 24
    seconds = seconds % 60
    minutes = minutes % 60
    hours = hours % 24
    return f"Day {days} - {hours:02}:{minutes:02}:{seconds:02}"

def wait(ms):
    start_time = time.time()
    end_time = start_time + ms / 1000
    while time.time() < end_time:
        pass
    return format_time(time.time() - start_time)

def wait_seconds(seconds):
    start_time = time.time()
    end_time = start_time + seconds
    while time.time() < end_time:
        pass
    return format_time(time.time() - start_time)

def wait_minutes(minutes):
    start_time = time.time()
    end_time = start_time + minutes * 60
    while time.time() < end_time:
        pass
    return format_time(time.time() - start_time)

def wait_hours(hours):
    start_time = time.time()
    end_time = start_time + hours * 3600
    while time.time() < end_time:
        pass
    return format_time(time.time() - start_time)

def ct(*args):
    current_time = time.localtime()
    result = []
    if 'time' in args:
        result.append(time.strftime("%H:%M:%S", current_time))
    if 'date' in args:
        result.append(time.strftime("%Y-%m-%d", current_time))
    if 'year' in args:
        result.append(time.strftime("%Y", current_time))
    if not result:
        result.append(time.strftime("%Y-%m-%d %H:%M:%S", current_time))
    return ", ".join(result)

def time_elapsed(start_time):
    return format_time(time.time() - start_time)



import os
import qrcode

def newqr(data, filename="newqr.png"):
    current_directory = os.path.dirname(os.path.realpath(__file__))
    filepath = os.path.join(current_directory, filename)

    qr = qrcode.QRCode(version=1, error_correction=qrcode.constants.ERROR_CORRECT_L, box_size=10, border=0)
    qr.add_data(data)
    qr.make(fit=True)
    qr.make_image(fill='black', back_color='white').save(filepath)

    print(f"QR code saved as {filepath}")



import os
import json
import psutil

# # # --- FIND/F/ALL --- # # #

def get_available_drives():
    drives = []
    partitions = psutil.disk_partitions()
    for partition in partitions:
        if partition.fstype:
            drives.append(partition.device)
    return drives

def find(file_name, max='all', min=1):
    if max == 0 or min == 0:
        return 'ERR_NOT_FOUND'

    try:
        max = int(max) if max != 'all' else float('inf')
        min = int(min) if min != 'all' else 1
    except ValueError:
        pass

    if min > max:
        return 'ERR_MIN_HIGHER'

    profile = os.environ.get('USERPROFILE')
    drives = get_available_drives()
    
    found_files = []
    
    for drive in drives:
        try:
            for root, dirs, files in os.walk(drive + '\\'):
                for file in files:
                    if file.lower() == file_name.lower():
                        found_files.append(os.path.join(root, file))
        except PermissionError:
            continue
        except FileNotFoundError:
            return 'ERR_HARDWARE_DISCONNECTED'
    
    if len(found_files) < min:
        return 'ERR_NOT_FOUND'
    
    if max == 'all':
        return json.dumps(found_files, indent=4)
    else:
        return json.dumps(found_files[:max], indent=4)

def findf(folder_name, max='all', min=1):
    if max == 0 or min == 0:
        return 'ERR_NOT_FOUND'

    try:
        max = int(max) if max != 'all' else float('inf')
        min = int(min) if min != 'all' else 1
    except ValueError:
        pass

    if min > max:
        return 'ERR_MIN_HIGHER'

    profile = os.environ.get('USERPROFILE')
    drives = get_available_drives()

    found_folders = []
    
    for drive in drives:
        try:
            for root, dirs, files in os.walk(drive + '\\'):
                for folder in dirs:
                    if folder.lower() == folder_name.lower():
                        found_folders.append(os.path.join(root, folder))
        except PermissionError:
            continue
        except FileNotFoundError:
            return 'ERR_HARDWARE_DISCONNECTED'
    
    if len(found_folders) < min:
        return 'ERR_NOT_FOUND'
    
    if max == 'all':
        return json.dumps(found_folders, indent=4)
    else:
        return json.dumps(found_folders[:max], indent=4)

def findall(criteria, max='all', min=1):
    if max == 0 or min == 0:
        return 'ERR_NOT_FOUND'

    try:
        max = int(max) if max != 'all' else float('inf')
        min = int(min) if min != 'all' else 1
    except ValueError:
        pass

    if min > max:
        return 'ERR_MIN_HIGHER'

    profile = os.environ.get('USERPROFILE')
    drives = get_available_drives()
    
    extensions = criteria.get('extensions', None)
    files = criteria.get('files', None)
    folders = criteria.get('folders', None)

    found_files = []
    found_folders = []

    for drive in drives:
        try:
            for root, dirs, files_in_root in os.walk(drive + '\\'):
                for file in files_in_root:
                    if files:
                        if any(file.lower() == f.lower() for f in files):
                            if extensions is None or any(file.lower().endswith(ext.lower()) for ext in extensions):
                                found_files.append(os.path.join(root, file))

                for folder in dirs:
                    if folders:
                        if any(folder.lower() == f.lower() for f in folders):
                            found_folders.append(os.path.join(root, folder))
        except PermissionError:
            continue
        except FileNotFoundError:
            return 'ERR_HARDWARE_DISCONNECTED'
    
    result = {}
    
    if found_files:
        result['files'] = found_files
    if found_folders:
        result['folders'] = found_folders
    
    if len(found_files) < min and len(found_folders) < min:
        return 'ERR_NOT_FOUND'
    
    if max == 'all':
        return json.dumps(result, indent=4) if result else 'ERR_NOT_FOUND'
    else:
        if len(found_files) < max or len(found_folders) < max:
            return json.dumps(result, indent=4) if result else 'ERR_NOT_FOUND'
        return json.dumps(result, indent=4) if result else 'ERR_NOT_FOUND'

# # # --- FIND/F/ALL --- # # #

# # # --- RM/CREATE --- # # #

SUCCESS_CREATE_FILE = "SUCCESS_CREATE_FILE"
ERR_CREATE_FILE = "ERR_CREATE_FILE"
SUCCESS_DELETE_FILE = "SUCCESS_DELETE_FILE"
ERR_DELETE_FILE = "ERR_DELETE_FILE"
ERR_NO_PERMS = "ERR_NO_PERMS"
ERR_FILE_NOT_FOUND = "ERR_FILE_NOT_FOUND"

def create(file_path, content=""):
    try:
        with open(file_path, 'w') as file:
            file.write(content)
        return SUCCESS_CREATE_FILE
    except PermissionError:
        return ERR_NO_PERMS
    except Exception:
        return ERR_CREATE_FILE

def rm(file_path):
    try:
        if os.path.exists(file_path):
            os.remove(file_path)
            return SUCCESS_DELETE_FILE
        else:
            return ERR_FILE_NOT_FOUND
    except PermissionError:
        return ERR_NO_PERMS
    except Exception:
        return ERR_DELETE_FILE

# # # --- RM/CREATE --- # # #


import win32api
import win32gui
import pyaudio
import wave
import numpy as np
import noisereduce as nr
import time

WM_APPCOMMAND = 0x319
APPCOMMAND_MICROPHONE_VOLUME_MUTE = 0x180000

SUCCESS_MIC_MUTE = "SUCCESS_MIC_MUTE"
SUCCESS_MIC_UNMUTE = "SUCCESS_MIC_UNMUTE"
ERR_MIC_MUTE = "ERR_MIC_MUTE"
ERR_MIC_UNMUTE = "ERR_MIC_UNMUTE"
SUCCESS_RECORDING = "SUCCESS_RECORDING"
ERR_RECORDING = "ERR_RECORDING"
ERR_NO_DEVICE = "ERR_NO_DEVICE"

def micmute():
    try:
        hwnd_active = win32gui.GetForegroundWindow()
        win32api.SendMessage(hwnd_active, WM_APPCOMMAND, None, APPCOMMAND_MICROPHONE_VOLUME_MUTE)
        return SUCCESS_MIC_MUTE
    except Exception:
        return ERR_MIC_MUTE

def micunmute():
    try:
        hwnd_active = win32gui.GetForegroundWindow()
        win32api.SendMessage(hwnd_active, WM_APPCOMMAND, None, APPCOMMAND_MICROPHONE_VOLUME_MUTE)
        return SUCCESS_MIC_UNMUTE
    except Exception:
        return ERR_MIC_UNMUTE

def list_audio_devices():
    p = pyaudio.PyAudio()
    print("Available audio devices:")
    for index in range(p.get_device_count()):
        device_info = p.get_device_info_by_index(index)
        print(f"Index {index}: {device_info['name']} (inputs: {device_info['maxInputChannels']})")
    p.terminate()

def get_system_audio_device_index():
    p = pyaudio.PyAudio()
    
    potential_devices = []
    
    for index in range(p.get_device_count()):
        device_info = p.get_device_info_by_index(index)
        if device_info['maxInputChannels'] > 0:
            name = device_info['name'].lower()
            
            if ('miks stereo' in name or 'stereo mix' in name or 
                'what u hear' in name or 'loopback' in name):
                potential_devices.append((index, 5))
            elif 'virtual' in name and 'cable' in name:
                potential_devices.append((index, 4))
            elif 'line' in name and device_info['maxInputChannels'] >= 2:
                potential_devices.append((index, 3))
            elif 'virtual' in name:
                potential_devices.append((index, 2))
            elif device_info['maxInputChannels'] >= 2:
                potential_devices.append((index, 1))
    
    potential_devices.sort(key=lambda x: x[1], reverse=True)
    
    if potential_devices:
        for device_index, priority in potential_devices:
            device_info = p.get_device_info_by_index(device_index)
            print(f"Trying system audio device: {device_info['name']} (Index: {device_index})")
            
            try:
                test_stream = p.open(
                    format=pyaudio.paInt16,
                    channels=2,
                    rate=44100,
                    input=True,
                    input_device_index=device_index,
                    frames_per_buffer=1024,
                    start=False
                )
                test_stream.close()
                print(f"Selected system audio device: {device_info['name']} (Index: {device_index})")
                p.terminate()
                return device_index
            except Exception as e:
                print(f"Cannot use device {device_index}: {e}")
    
    print("No working system audio device found")
    p.terminate()
    return None

def get_mic_input_device_index():
    p = pyaudio.PyAudio()
    
    potential_devices = []
    
    for index in range(p.get_device_count()):
        device_info = p.get_device_info_by_index(index)
        if device_info['maxInputChannels'] > 0:
            name = device_info['name'].lower()
            
            if 'fifine' in name:
                potential_devices.append((index, 5))
            elif 'microphone' in name:
                potential_devices.append((index, 4))
            elif 'mic' in name:
                potential_devices.append((index, 3))
            elif 'input' in name:
                potential_devices.append((index, 2))
            else:
                potential_devices.append((index, 1))
    
    potential_devices.sort(key=lambda x: x[1], reverse=True)
    
    if potential_devices:
        for device_index, priority in potential_devices:
            device_info = p.get_device_info_by_index(device_index)
            print(f"Trying microphone device: {device_info['name']} (Index: {device_index})")
            
            try:
                test_stream = p.open(
                    format=pyaudio.paInt16,
                    channels=1,
                    rate=44100,
                    input=True,
                    input_device_index=device_index,
                    frames_per_buffer=1024,
                    start=False
                )
                test_stream.close()
                print(f"Selected microphone device: {device_info['name']} (Index: {device_index})")
                p.terminate()
                return device_index
            except Exception as e:
                print(f"Cannot use device {device_index}: {e}")

    print("No working microphone device found")
    p.terminate()
    return None

def record(time_ms, path, noise_removal=False, progress_callback=None):
    try:
        p = pyaudio.PyAudio()
        rate = 44100
        channels = 1
        format = pyaudio.paInt16
        frames_per_buffer = 1024

        input_device_index = get_mic_input_device_index()
        if input_device_index is None:
            return ERR_NO_DEVICE

        stream = p.open(format=format,
                        channels=channels,
                        rate=rate,
                        input=True,
                        input_device_index=input_device_index,
                        frames_per_buffer=frames_per_buffer)

        frames = []
        time_seconds = time_ms / 1000
        total_frames = int(rate / frames_per_buffer * time_seconds)

        for i in range(total_frames):
            try:
                data = stream.read(frames_per_buffer)
                frames.append(data)
            except Exception as e:
                return f'WARN_DURING_RECORDING\n{e}'

            if progress_callback:
                progress = int((i / total_frames) * 100)
                progress_callback(progress)

        stream.stop_stream()
        stream.close()

        if not frames:
            return f'ERR_RECORDING\n{e}'

        audio_data = b''.join(frames)
        audio_np = np.frombuffer(audio_data, dtype=np.int16)
        
        if noise_removal and len(audio_np) > 0:
            try:
                # Zakładamy, że pierwsze 0.25 sekundy to sam szum
                noise_sample = audio_np[:int(0.25 * rate)]
                audio_np = nr.reduce_noise(y=audio_np, y_noise=noise_sample, sr=rate, prop_decrease=0.9999)

            except Exception as e:
                print(f"Noise reduction failed: {e}")
        
        if len(audio_np) > 0 and np.max(np.abs(audio_np)) > 0:
            audio_np = np.int16(audio_np / np.max(np.abs(audio_np)) * 32767)

        with wave.open(path, 'wb') as wf:
            wf.setnchannels(channels)
            wf.setsampwidth(p.get_sample_size(format))
            wf.setframerate(rate)
            wf.writeframes(audio_np.tobytes())

        p.terminate()
        return SUCCESS_RECORDING
    except Exception as e:
        print(f"Error during recording: {e}")
        return ERR_RECORDING


import math

def add(a, b):
    return a + b

def sub(a, b):
    return a - b

def mult(a, b):
    return a * b

def div(a, b):
    if b == 0:
        raise ValueError("Cannot divide by zero")
    return a / b

def power(a, b):
    return a ** b

def sqrt(a):
    if a < 0:
        raise ValueError("Cannot compute the square root of a negative number")
    return math.sqrt(a)

def abs_value(a):
    return abs(a)

def log(a):
    if a <= 0:
        raise ValueError("Logarithm undefined for non-positive numbers")
    return math.log(a)

def log10(a):
    if a <= 0:
        raise ValueError("Logarithm undefined for non-positive numbers")
    return math.log10(a)

def sin(a):
    return math.sin(a)

def cos(a):
    return math.cos(a)

def tan(a):
    return math.tan(a)

def atan(a):
    return math.atan(a)

def nth_root(a, n):
    if n == 0:
        raise ValueError("Cannot compute the 0th root")
    return a ** (1/n)

def factorial(a):
    if a < 0:
        raise ValueError("Factorial is undefined for negative numbers")
    return math.factorial(a)

def exp(a):
    return math.exp(a)

def pi():
    return math.pi

def e():
    return math.e

def sin_deg(a):
    return math.sin(math.radians(a))

def cos_deg(a):
    return math.cos(math.radians(a))

def tan_deg(a):
    return math.tan(math.radians(a))

def derivative(f, x, h=1e-5):
    return (f(x + h) - f(x)) / h

def integral(f, a, b, n=1000):
    h = (b - a) / n
    return sum(f(a + i * h) * h for i in range(n))

def gcd(a, b):
    while b:
        a, b = b, a % b
    return a

def lcm(a, b):
    return abs(a * b) // gcd(a, b)

def combination(n, r):
    return factorial(n) // (factorial(r) * factorial(n - r))

def permutation(n, r):
    return factorial(n) // factorial(n - r)



import sys

# Only for windows!
use_audio = sys.platform == "win32"

if use_audio:
    from ctypes import cast, POINTER
    from comtypes import CLSCTX_ALL
    from pycaw.pycaw import AudioUtilities, IAudioEndpointVolume

    def mute():
        try:
            devices = AudioUtilities.GetSpeakers()
            interface = devices.Activate(IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
            volume = cast(interface, POINTER(IAudioEndpointVolume))
            volume.SetMute(1, None)
            return "SUCCESS_MUTE"
        except Exception as e:
            return f"ERR_MUTE: {str(e)}"

    def unmute():
        try:
            devices = AudioUtilities.GetSpeakers()
            interface = devices.Activate(IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
            volume = cast(interface, POINTER(IAudioEndpointVolume))
            volume.SetMute(0, None)
            return "SUCCESS_UNMUTE"
        except Exception as e:
            return f"ERR_UNMUTE: {str(e)}"

else:
    print("Audio control features are disabled on this platform.")